/**
 * Benchmark of digital IIR filters.
 */

#define __STDC_LIMIT_MACROS
#define __STDC_FORMAT_MACROS

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <inttypes.h>
#include <unistd.h>
#include <sys/time.h>


/**
 * Number of controlled channels.
 * This is also the size of the (square) matrices.
 * Do not change; several parts of the code make assumptions about this value.
 */
#define NCHANNEL        6

/**
 * Number of channels that are coupled through input/output matrices.
 * The non-diagonal elements of the matrix are only taken into account
 * for the first NMBCHANNEL channels. Other channels are assumed to be
 * independent and only use the diagonal elements of the matrix.
 * Do not change; it is also hardcoded in the matrix multiplication code.
 */
#define NMBCHANNEL      6

/**
 * Number of IIR filter sections per channel.
 * 1 section for PID controller; 5 sections for notch filter.
 */
#define IIRSECTIONS     5


/**
 * Coefficients of one 2nd order IIR filter section.
 *
 * To avoid overflow, the sum of absolute values of these 5 coefficients
 * must be less than 2**32; otherwise wraparound may cause incorrect results.
 */
struct iirfilter_section_coeff {
    int32_t         b0, b1, b2, a1, a2;
    int32_t         scalebits;
} __attribute__ ((packed)) __attribute__ ((aligned (4)));

typedef struct iirfilter_section_coeff iirfilter_coeff[IIRSECTIONS];


/**
 * Internal state variables of the PID controller (for a single channel).
 */
typedef struct {
    int32_t         prev;
    int32_t         accum;
} pid_state;


/**
 * Internal state variables of the IIR filter (for a single channel).
 * The output state variables of one filter section are shared with the
 * input state variables of the next section.
 */
typedef int32_t iirfilter_state[IIRSECTIONS+1][2];


/**
 * Internal state valiables of the random number generator.
 */
typedef struct {
    uint32_t        s1, s2, s3;
} random_state;


/**
 * Coefficients of the full control loop.
 *
 * To avoid wraparound-on-overflow, the following conditions must
 * be satisfied:
 *  * In each row of inmatrix and outmatrix, the sum of absolute values
 *    of the elements of that row must be less than 2**32.
 *  * Each PID setpoint must have an absolute value less than 2**31.
 *  * In each PID filter, the sum of absolute values of kp, ki and kd
 *    must be less than 2**32.
 *  * In each IIR filter section, the sum of absolute values of the
 *    five filter coefficients must be less than 2**32.
 *
 * There are separate scaling factors for the input matrix,
 * output matrix, each PID controller and each IIR filter section.
 * Scaling factors are given as the number of bits to shift to the right
 * at the output of the sub-routine. Each scaling factor must be at least 1
 * and at most 31.
 */
struct controller_coeff {
    int32_t         inmatrix[NCHANNEL*NCHANNEL];
    int32_t         outmatrix[NCHANNEL*NCHANNEL];
    int32_t         inmatscalebits;
    int32_t         outmatscalebits;        
    int32_t         setpoint[NCHANNEL];
    iirfilter_coeff filter[NCHANNEL];
    int32_t         noisegain[NCHANNEL];
};


/**
 * State for the full control loop.
 */
struct controller_state {
    pid_state       pid[NCHANNEL];
    iirfilter_state filter[NCHANNEL];
    random_state    rand;
    int32_t         noise;
};


/**
 * Reset all state variables to zero and reseed the noise generator.
 */
void algorithm_reset(const struct controller_coeff *coeff,
                     struct controller_state *state,
                     uint32_t seed);


/**
 * Compute the next iteration through the control loop.
 *
 * @param  coeff    Coefficients.
 * @param  state    State variables; updated in-place.
 * @param  invec    Array of NCHANNEL input values.
 * @param  outvec   Room for NCHANNEL outut values; may overlap with invec.
 * @note   invec and outvec may overlap;
 *         elements from invec are used before updating outvec.
 */
void algorithm_run(const struct controller_coeff *coeff,
                   struct controller_state *state,
                   const int32_t *invec,
                   int32_t *outvec);


/**
 * Shift x to the right by shiftcnt bits (1 <= shiftcnt <= 31) and
 * return the least significant 32 bits of the result.
 *
 * It could be written as (int32_t)(x >> shiftcnt), but in that case
 * the compiler calls a slow library subroutine to do a full 64-bit shift.
 * By writing it out like this, we take advantage of the fact that shiftcnt
 * is between 1 and 31 and only a few machine instructions are needed.
 */
static inline int32_t shiftright_64to32(int64_t x, unsigned int shiftcnt)
{
    uint32_t xl = (uint32_t)x;  /* note: unsigned type of xl is essential */
    int32_t  xh = (int32_t)(x >> 32);
    return (xl >> shiftcnt) | (xh << (32 - shiftcnt));
}


/**
 * Shift x to the right by shiftcnt bits (1 <= shiftcnt <= 31).
 * If the result does not fit in a signed 32-bit integer, return
 * either INT32_MIN or INT32_MAX depending on the sign of x.
 */
static inline int32_t scaledown_saturated(int64_t x, unsigned int shiftcnt)
{
    /*
     * t<31 downto 1> = high bits of x which don't fit in result
     * t<0>           = most significant bit that still fits in result
     * If the shifted value fits in y without losing bits, t will be
     * either 0 or -1; otherwise overflow has occurred.
     */
    int32_t t = ((int32_t)(x >> 32)) >> (shiftcnt - 1);
    return (t < -1) ? INT32_MIN :
           (t > 0) ? INT32_MAX : shiftright_64to32(x, shiftcnt);
}


/**
 * Return the sum of a and b; if the absolute value of the result exceeds
 * m (or if the result does not fit in an int32_t), return either m
 * or -m depending ot the true sign of the result.
 * Note: m must be positive.
 */
static inline int32_t add_saturated(int32_t a, int32_t b, int32_t m)
{
    return (a >= 0 && m - a < b) ? m :
           (a < 0 && -m - a > b) ? (-m) : (a + b);
}


/**
 * Maximally equidistributed combined Tausworthe generator,
 * taken from the GNU Scientific Library.
 */
static inline uint32_t random_gen(random_state *state)
{
    #define TAUSWORTHE(s,a,b,c,d) ( ((s & c) << d) ^ (((s << a) ^ s) >> b) )
    state->s1 = TAUSWORTHE(state->s1, 13, 19, 4294967294UL, 12);
    state->s2 = TAUSWORTHE(state->s2,  2, 25, 4294967288UL,  4);
    state->s3 = TAUSWORTHE(state->s3,  3, 11, 4294967280UL, 17);
    #undef TAUSWORTHE
    return (state->s1 ^ state->s2 ^ state->s3);
}


/** Seed the random generator. */
static void random_seed(random_state *state, uint32_t seed)
{
    int i;
    if (seed == 0)
        seed = 1;
    #define LCG(n) (69069 * n)
    state->s1 = LCG(seed);
    if (state->s1 < 2) state->s1 += 2UL;
    state->s2 = LCG(state->s1);
    if (state->s2 < 8) state->s2 += 8UL;
    state->s3 = LCG(state->s2);
    if (state->s3 < 16) state->s3 += 16UL;
    #undef LCG
    for (i = 0; i < 6; i++)
        random_gen(state);
}


/**
 * Matrix-vector multiplication; 32x32 mul, 64-bit accum, 32-bit result.
 *
 * Note: invec and outvec may overlap;
 *       elements from invec are used before updating outvec.
 */
static void matrixmul(const int32_t *matrix,
                      const int32_t *invec,
                      int32_t *outvec,
                      unsigned int scalebits)
{
#if (NMBCHANNEL != 5 && NMBCHANNEL != 6)
#error "This code only works for NMBCHANNEL == 5 or NMBCHANNEL == 6"
#endif
    unsigned int i;
    register int32_t in0 = invec[0];
    register int32_t in1 = invec[1];
    register int32_t in2 = invec[2];
    register int32_t in3 = invec[3];
    register int32_t in4 = invec[4];
#if (NMBCHANNEL == 6)
    register int32_t in5 = invec[5];
#endif
    for (i = 0; i < NMBCHANNEL; i++) {
        int64_t t = 0;
        t += (int64_t)(matrix[0]) * (int64_t)in0;
        t += (int64_t)(matrix[1]) * (int64_t)in1;
        t += (int64_t)(matrix[2]) * (int64_t)in2;
        t += (int64_t)(matrix[3]) * (int64_t)in3;
        t += (int64_t)(matrix[4]) * (int64_t)in4;
#if (NMBCHANNEL == 6)
        t += (int64_t)(matrix[5]) * (int64_t)in5;
#endif
        matrix += NCHANNEL;
        outvec[i] = scaledown_saturated(t, scalebits);
    }
    for (i = NMBCHANNEL; i < NCHANNEL; i++) {
        int64_t t = (int64_t)(matrix[i]) * (int64_t)invec[i];
        outvec[i] = scaledown_saturated(t, scalebits);
        matrix += NCHANNEL;
    }
}


/**
 * Compute the IIR filter chain for a single channel.
 */
static int32_t iirfilter(int32_t x,
                         const iirfilter_coeff *coeff,
                         iirfilter_state *state)
{
    int32_t x1, x2;
    unsigned int j;

    /* Load input state variables of first section. */
    x1 = (*state)[0][0];
    x2 = (*state)[0][1];

    for (j = 0; j < IIRSECTIONS; j++) {
        /* Compute one section. */
        int64_t acc;
        (*state)[j][0] = x;
        (*state)[j][1] = x1;
        acc =  (int64_t)((*coeff)[j].b0) * (int64_t)x;
        acc += (int64_t)((*coeff)[j].b1) * (int64_t)x1;
        acc += (int64_t)((*coeff)[j].b2) * (int64_t)x2;
        x1 = (*state)[j+1][0];
        x2 = (*state)[j+1][1];
        acc -= (int64_t)((*coeff)[j].a1) * (int64_t)x1;
        acc -= (int64_t)((*coeff)[j].a2) * (int64_t)x2;
        x = scaledown_saturated(acc, (*coeff)[j].scalebits);
    }

    /* Update output state variables of last section. */
    (*state)[j][0] = x;
    (*state)[j][1] = x1;

    return x;
}


/**
 * Reset all state variables to zero and reseed the noise generator.
 */
void algorithm_reset(const struct controller_coeff *coeff,
                     struct controller_state *state,
                     uint32_t seed)
{
    memset(state, 0, sizeof(*state));
    random_seed(&state->rand, seed);
}


/**
 * Compute the next iteration through the control loop.
 *
 * Note: invec and outvec may overlap;
 *       elements from invec are used before updating outvec.
 */
void algorithm_run(const struct controller_coeff *coeff,
                   struct controller_state *state,
                   const int32_t *invec,
                   int32_t *outvec)
{
    unsigned int i;

    /* add setpoints */
    for (i = 0; i < NCHANNEL; i++) {
        outvec[i] = add_saturated(invec[i], coeff->setpoint[i], INT32_MAX);
    }

    /* input matrix */
    matrixmul(coeff->inmatrix, outvec, outvec, coeff->inmatscalebits);

    /* IIR filters */
    for (i = 0; i < NCHANNEL; i++) {
        outvec[i] = iirfilter(outvec[i], &coeff->filter[i], &state->filter[i]);
    }

    /* output matrix */
    matrixmul(coeff->outmatrix, outvec, outvec, coeff->outmatscalebits);

    /* noise injection */
    int32_t z = (int32_t) (random_gen(&state->rand) & 0xffff0000);
    state->noise = z;
    for (i = 0; i < NCHANNEL; i++) {
        int64_t tt = (int64_t)coeff->noisegain[i] * (int64_t)z;
        int32_t t = (int32_t)(tt >> 31);
        outvec[i] = add_saturated(outvec[i], t, INT32_MAX);
    }
}


#define NLOOP 100000
int main(void)
{
    static const int32_t imat[6*6] = {
    16777216 ,   1247628 ,   2928275 ,  -5236662 ,   -568103 ,        0 ,
    -2379679 ,  16777216 ,  -1933737 ,     78433 ,  -2303950 ,        0 ,
     -320790 ,  -1485891 ,  16777216 ,  12509362 ,  -3122743 ,        0 ,
      937400 , -21022920 , -14560161 ,  16777216 ,  -9757990 ,        0 ,
     6933832 ,   8122005 ,  -2598756 ,   3380607 ,  16777216 ,        0 ,
           0 ,         0 ,         0 ,         0 ,         0 ,  8500000 };
    static const int32_t omat[6*6] = {
    16777216 ,   4820586 ,  -9257102 ,  -2611994 ,    341884 ,        0 ,
     5759860 ,  16777216 ,  -5181059 , -11301619 ,   -135281 ,        0 ,
    -4054869 ,  -6503778 ,  16777216 ,  -5165646 ,  -2588438 ,        0 ,
   -17196811 ,    483288 ,  25626899 ,  16777216 ,   8923722 ,        0 ,
   -12256442 , -15465357 ,   5605301 ,   3274406 ,  16777216 ,        0 ,
           0 ,         0 ,         0 ,         0 ,         0 , 30000000 };
    static const iirfilter_coeff filter = {
        {  16777215 ,   16224076 ,         0 , -507169471 , 242761558 , 28 },
        { 268435455 , -515535726 , 268233748 , -500835577 , 253533590 , 28 },
        { 268435455 , -519371151 , 268285781 , -506008655 , 254923308 , 28 },
        { 268435455 , -526199101 , 268364315 , -515690069 , 257855226 , 28 },
        {  32760447 ,  -64521289 ,  32754614 , -519447013 , 259154466 , 28 } };
    static struct controller_coeff coeff;
    static struct controller_state state;
    static int32_t tvec[NLOOP][NCHANNEL];
    random_state rstate;
    struct timeval tvs, tve;
    unsigned int i, j;

    printf("\nOdelco benchmark\n");

    /* Set up coefficients */
    for (i = 0; i < NCHANNEL; i++)
        coeff.setpoint[i] = -5000 + 1100 * i;
    memcpy(coeff.inmatrix, imat, sizeof(coeff.inmatrix));
    memcpy(coeff.outmatrix, omat, sizeof(coeff.outmatrix));
    coeff.inmatscalebits = 24;
    coeff.outmatscalebits = 24;
    for (i = 0; i < NCHANNEL; i++) {
        memcpy(coeff.filter[i], filter, sizeof(coeff.filter[i]));
        for (j = 0; j < IIRSECTIONS; j++) {
            coeff.filter[i][j].b0 += i * (j + 1);
            coeff.filter[i][j].b1 += i * (j + 10);
            coeff.filter[i][j].b2 += i * (j + 20);
            coeff.filter[i][j].a1 += i * (j + 30);
            coeff.filter[i][j].a2 += i * (j + 40);
        }
    }
    for (i = 0; i < NCHANNEL; i++)
        coeff.noisegain[i] = i * 100;

    /* Create random input data */
    random_seed(&rstate, 1);
    for (i = 0; i < NLOOP; i++)
        for (j = 0; j < NCHANNEL; j++)
            tvec[i][j] = (int32_t)random_gen(&rstate);

    /* Run algorithm */
    printf("Timing %d cycles:\n", NLOOP);
    gettimeofday(&tvs, NULL);
    algorithm_reset(&coeff, &state, 1);
    for (i = 1; i < NLOOP; i++)
        algorithm_run(&coeff, &state, tvec[i], tvec[i]);
    gettimeofday(&tve, NULL);

    /* Print timing info */
    printf("\n%d cycles in %3ld.%06ld s\n", NLOOP,
           tve.tv_sec - tvs.tv_sec - ((tve.tv_usec < tvs.tv_usec) ? 1 : 0),
           tve.tv_usec - tvs.tv_usec + ((tve.tv_usec < tvs.tv_usec) ? 1000000 : 0));

    /* Verify output data (checksum) */
    {
        uint32_t sum1 = 0xffffffff, sum2 = 0xffffffff;
        for (i = 0; i < NLOOP; i++) {
            for (j = 0; j < NCHANNEL; j++) {
                uint64_t t = (uint64_t)sum1 + (uint32_t)tvec[i][j];
                sum1 = (uint32_t)t + (uint32_t)(t >> 32);
                uint64_t u = (uint64_t)sum2 + sum1;
                sum2 = (uint32_t)u + (uint32_t)(u >> 32);
            }
        }
        printf("checksum = %08"PRIx32"%08"PRIx32" (%s)\n",
               sum2, sum1,
               (sum2 == 0x12b6184b && sum1 == 0x77b7c01b) ? "ok" : "BAD");
    }

    fflush(stdout);
    return 0;
}

