"""
Find LEON3 pipeline hazards in a SPARC assembler file.

This tool has a very limited understanding of assembler syntax, instruction
flow and the LEON3 pipeline. It understands assembler syntax as emitted
by GCC. Not all instruction types are implemented.

Conditional branches are followed (both ways), but loops are not recognized.
Each instruction is counted only once, whether or not it occurs in a loop.
Calls are not tracked; each subroutine should be analyzed separately.

The tool outputs the following information for each subroutine in the code:
  Intrinsic cost:   Total number of intrinsic execution cycles.
  Data stalls:      Extra cycles due to data dependency (after LOAD or MUL).
  Branch stalls:    Extra cycles due to conditional branch depending on
                    condition codes.
  Store buffer:     Extra cycles due to collision between pending store
                    and new load (assuming load causes cache miss).

This tool is only useful for identifying certain types of pipeline stalls.
It can not be used to predict execution time.

Usage: pipehazard.py asmfile.s
"""

import sys, re

MUL_COST = 5                # intrinsic cost of MUL instruction
STORE_FLUSH_COST = 4        # cost of flushing a pending store buffer

def parsefile(fname):

    insns = [ ]
    syms = { }

    pendingsym = None

    f = file(fname)

    i = 0
    while 1:

        if pendingsym and pendingsym[1] < len(insns):
            syms[pendingsym[0]] = pendingsym[1]
            pendingsym = None

        s = f.readline()
        if not s:
            break
        i += 1

        m = re.match(r'([A-Za-z0-9._]+):', s)
        if m:
            sym = m.group(1)
            if sym.startswith('.'):
                syms[sym] = len(insns)
            else:
                pendingsym = (sym, len(insns))
            continue

        m = re.match(r'\s*(save|restore|unimp)\s+', s)
        if m:
            if m.group(1) == 'unimp':
                print "ignoring 'unimp' instruction on line %d" % i
            insns.append( (i, '', m.group(1)) )
            continue

        m = re.match(r'\s*(nop)\s+', s)
        if m:
            insns.append( (i, 'alu', m.group(1), [ ], 0) )
            continue

        m = re.match(r'\s*(jmp|jmp|ret|retl)\s+', s)
        if m:
            insns.append( (i, 'jmp', m.group(1)) )
            continue

        m = re.match(r'\s*(call)\s+', s)
        if m:
            insns.append( (i, 'call', m.group(1)) )
            continue

        m = re.match(r'\s*(b|ba|be|beq|bne|bl|ble|bg|bge|bleu|bgeu|bgu|blu|bneg|bpos)(,a)?\s+([A-Za-z0-9._]+)\s+', s)
        if m:
            op = m.group(1)
            if m.group(1) in ('b', 'ba'):
                ty = 'branch'
            else:
                ty = 'bcond'
            if m.group(2):
                op += m.group(2)
            insns.append( (i, ty, op, m.group(3)) )
            continue

        m = re.match(r'\s*(ld|ldub|lduh|ldsb|ldsh|ldd)\s+\[([^]]+)\]\s*,\s*(%[a-z][0-9])\s+', s)
        if m:
            regs = [ reg for reg in re.split(r'[-+]', m.group(2)) if reg.startswith('%') ]
            insns.append( (i, 'ld', m.group(1), regs, m.group(3)) )
            continue

        m = re.match(r'\s*(st|stb|sth|std)\s+(%[a-z][0-9])\s*,\s*\[([^]]+)\]\s+', s)
        if m:
            regs = [ reg for reg in re.split(r'[-+]', m.group(3)) if reg.startswith('%') ]
            insns.append( (i, 'st', m.group(1), regs, m.group(2)) )
            continue

        m = re.match(r'\s*(mov|rd)\s+([^,]+)\s*,\s*(%[a-z][0-9])\s+', s)
        if m:
            regs = [ ]
            if re.match(r'%[a-z][0-9]$', m.group(2)):
                regs.append(m.group(2))
            insns.append( (i, 'mov', m.group(1), regs, m.group(3)) )
            continue

        m = re.match(r'\s*(set|sethi)\s+([^,]+)\s*,\s*(%[a-z][0-9])\s+', s)
        if m:
            insns.append( (i, 'mov', m.group(1), [ ], m.group(3)) )
            continue

        m = re.match(r'\s*(wr)\s+(%[a-z][0-9])\s*,\s*(-?[0-9]+)\s*,\s*(%asr[0-9]+|%y)\s+', s)
        if m:
            insns.append( (i, 'mov', m.group(1), [ m.group(2) ], m.group(4)) )
            continue

        m = re.match(r'\s*(cmp)\s+(%[a-z][0-9])\s*,\s*(%[a-z][0-9])?', s)
        if m:
            regs = [ m.group(2) ]
            if m.group(3):
                regs.append(m.group(3))
            insns.append( (i, 'alucc', m.group(1), regs, 0) )
            continue

        m = re.match(r'\s*(sll|srl|sra|xor|xorcc|xnor|xnorcc|or|orcc|and|andcc|andn|andncc|addx?|addx?cc|subx?|subx?cc)\s+(%[a-z][0-9]|%fp)\s*,\s*([^,]+)\s*,\s*(%[a-z][0-9])\s+', s)
        if m:
            ty = 'alu'
            if m.group(1).endswith('cc'):
                ty += 'cc'
            regs = [ m.group(2) ]
            if re.match(r'%[a-z][0-9]$', m.group(3)):
                regs.append(m.group(3))
            insns.append( (i, ty, m.group(1), regs, m.group(4)) )
            continue

        m = re.match(r'\s*(udiv|sdiv|udivcc|sdivcc)\s+(%[a-z][0-9])\s*,\s*([^,]+)\s*,\s*(%[a-z][0-9])\s+', s)
        if m:
            ty = 'div'
            regs = [ m.group(2) ]
            if re.match(r'%[a-z][0-9]$', m.group(3)):
                regs.append(m.group(3))
            insns.append( (i, ty, m.group(1), regs, m.group(4)) )
            continue

        m = re.match(r'\s*(umul|smul|umulcc|smulcc)\s+(%[a-z][0-9])\s*,\s*([^,]+)\s*,\s*(%[a-z][0-9])\s+', s)
        if m:
            ty = 'mul'
            regs = [ m.group(2) ]
            if re.match(r'%[a-z][0-9]$', m.group(3)):
                regs.append(m.group(3))
            insns.append( (i, ty, m.group(1), regs, m.group(4)) )
            continue

        m = re.match(r'\s*\.', s)               # pseudo statement
        if m:
            continue

        m = re.match(r'\s*\S+\s+=\s+\S+', s)    # symbol assignment
        if m:
            continue

        print >>sys.stderr, "Unmatched statement '%s'" % s.strip()
        sys.exit(1)

    f.close()

    return insns, syms


def findhazards(insns, syms, startsym):

    def testhazard(pc, npc):
        if npc is not None:
            insn  = insns[pc]
            ninsn = insns[npc]
            if ( insn[1] in ('ld', 'mul', 'div') and 
                 ninsn[1] in ('ld', 'st', 'alu', 'alucc', 'mul', 'div') ):
                outreg1 = insn[4]
                inregs2 = ninsn[3]
                if outreg1 in inregs2:
                    print '1-cycle data stall,   line %4d:' % insn[0], insn[2:], '>', ninsn[2:]
                    return 1
        return 0

    print "**** Scanning '%s'" % startsym

    ninsn   = 0             # number of (reachable) instructions
    nload   = 0             # number of load instructions
    nstore  = 0             # number of store instructions
    nbranch = 0             # number of conditional branch instructions

    cost    = 0             # total intrinsic cost
    datahz  = 0             # total data dependency hazards
    branchhz = 0            # total branch hazards
    storehz = 0             # total pending-store hazards

    # Scan instruction flow graph.
    states = { }
    queue  = [ ]
    nextinsn = { }
    pc = syms[startsym]
    npc = pc + 1

    while 1:

        if (pc is None) or (pc, npc) in states:
            if not queue:
                break
            (pc, npc) = queue.pop()
        states[(pc, npc)] = 1

        if (npc is not None) and (npc + 1 < len(insns)):
            nnpc = npc + 1
        else:
            nnpc = None

        insn = insns[pc]
        ninsn += 1
        # Note: instructions which occur in a conditional branch slot are counted double.

        inreg = None
        outreg = None
        label = None
        if len(insn) ==  5:
            (linenr, ty, op, inreg, outreg) = insn
        elif len(insn) == 4:
            (linenr, ty, op, label) = insn
        else:
            (linenr, ty, op) = insn

        if ty == 'jmp':
            c = 3
            nnpc = None

        elif ty == 'call':
            c = 1

        elif ty == 'branch':
            c = 1
            nnpc = syms[label]
            if op.endswith(',a'):
                npc = nnpc
                nnpc = npc + 1
                c += 1

        elif ty == 'bcond':
            nbranch += 1
            c = 1
            newst = (npc, syms[label])
            if newst not in states:
                states[newst] = 1
                queue.append(newst)
            if op.endswith(',a') and (npc is not None):
                npc = nnpc
                nnpc = npc + 1
                c += 1

        elif ty == 'ld':
            nload += 1
            c = 1
            if op == 'ldd':
                c += 1
            datahz += testhazard(pc, npc)

        elif ty == 'st':
            nstore += 1
            c = 2
            if op == 'std':
                c += 1

        elif ty == 'mov':
            c = 1

        elif ty == 'alu':
            c = 1

        elif ty == 'alucc':
            c = 1
            if (npc is not None) and (insns[npc][1] == 'bcond'):
                print '2-cycle branch stall, line %4d:' % linenr, insn[2:], '>', insns[npc][2:]
                branchhz += 2
            elif (nnpc is not None) and (insns[nnpc][1] == 'bcond') and (insns[npc][1] not in ('st', 'mul', 'div')):
                print '1-cycle branch stall, line %4d:' % linenr, insn[2:], '>', insns[nnpc][2:]
                branchhz += 1

        elif ty == 'mul':
            c = MUL_COST
            datahz += testhazard(pc, npc)

        elif ty == 'div':
            c = 35
            datahz += testhazard(pc, npc)

        elif ty == '':
            c = 0

        else:
            print >>sys.stderr, 'Unknown instruction pattern', insn

        cost += c
        if pc not in nextinsn:
            nextinsn[pc] = [ ]
        if npc is not None:
            nextinsn[pc].append( (npc, c) )

        pc  = npc
        npc = nnpc

    # Analyze pending-store hazards.
    store_pending = { }
    queue = [ syms[startsym] ]
    while queue:
        pc = queue.pop()
        insn = insns[pc]
        ty = insn[1]
        for (npc, c) in nextinsn[pc]:
            if ty == 'st':
                p = STORE_FLUSH_COST
            else:
                p = store_pending.get(pc, 0) - c
            p = max(0, p)
            if (npc not in store_pending) or (p > store_pending[npc]):
                store_pending[npc] = p
                if npc not in queue:
                    queue.append(npc)
    for (pc, p) in store_pending.iteritems():
        insn = insns[pc]
        ty = insn[1]
        if ty == 'ld':
            # If this load hits a cache miss, there is an extra penalty
            # because pending writes must be flushed first.
            storehz += p

    print 'Instructions:             %4d' % ninsn
    print '  loads:                  %4d' % nload
    print '  stores:                 %4d' % nstore
    print '  conditional branches:   %4d' % nbranch
    print 'Total cost:               %4d' % (cost + datahz + branchhz + storehz)
    print '  intrinsic cost:         %4d' % cost
    print '  data stalls:            %4d' % datahz
    print '  branch stalls:          %4d' % branchhz
    print '  store buffer stalls:    %4d' % storehz
    print


def main():

    if len(sys.argv) != 2:
        print >>sys.stderr, __doc__
        sys.exit(1)

    insns, syms = parsefile(sys.argv[1])

    globals = [ (pc, sym) for (sym, pc) in syms.iteritems() if not sym.startswith('.') ]
    globals.sort()
    for (pc, sym) in globals:
        findhazards(insns, syms, sym)


main()
