#!/usr/bin/env perl

use strict;
use warnings;

sub emit_size {
    my ($nvar, $nclause) = @_;
    print "static const int nvar = $nvar ;\n";
    print "static const int nclause = $nclause ;\n";
}

sub emit_start_clauses {
    print "static const int clauses[] = {\n";
}

sub emit_end_clauses {
    print "0 };\n";
}

my $nlit = 0;
sub emit_lit {
    print $_[0] . ", ";
    $nlit++;
    print "\n" if (($nlit & 7) == 0);
}

sub parse_dimacs {
    my ($nvar, $nclause);
    my $cnt = 0;
    while (<STDIN>) {
        next if (/^\s*$/);
        next if (/^c/);
        if (/^p cnf ([0-9]+) ([0-9]+)/) {
            die if (defined($nvar));
            $nvar    = $1;
            $nclause = $2;
            emit_size($nvar, $nclause);
            emit_start_clauses();
            next;
        }
        die if (!defined($nvar));
        $cnt++;
        foreach my $t (split()) {
            emit_lit($t);
            last if ($t == 0);
        }
    }
    emit_end_clauses();
    die if ($cnt != $nclause);
}

print <<'EOF';
#define __STDC_LIMIT_MACROS
#define __STDC_FORMAT_MACROS

#include <stdint.h>
#include <inttypes.h>
#include <assert.h>
#include <errno.h>
#include <sys/time.h>
#include "core/Solver.h"
#include "core/Solver.cc"

using namespace Minisat;

EOF

parse_dimacs();

print <<'EOF';

void printStats(Solver& solver)
{
    printf("restarts              : %"PRIu64"\n", solver.starts);
    printf("conflicts             : %-12"PRIu64" \n", solver.conflicts);
    printf("decisions             : %-12"PRIu64"   (%4.2f %% random)\n", solver.decisions, (float)solver.rnd_decisions*100 / (float)solver.decisions);
    printf("propagations          : %-12"PRIu64" \n", solver.propagations);
    printf("conflict literals     : %-12"PRIu64"   (%4.2f %% deleted)\n", solver.tot_literals, (solver.max_literals - solver.tot_literals)*100 / (double)solver.max_literals);
}

static Solver* solver;

int main(void)
{
    struct timeval tstart, tend;
    Solver S;
    S.verbosity = 0;
    solver = &S;

    printf("\nSatisfiability benchmark\n");

    gettimeofday(&tstart, NULL);

    int j = 0;
    for (int i = 0; i < nclause; i++) {
        vec<Lit> lits;
        while (1) {
            int parsed_lit = clauses[j++];
            if (parsed_lit == 0) break;
            int var = abs(parsed_lit)-1;
            while (var >= S.nVars()) S.newVar();
            lits.push( (parsed_lit > 0) ? mkLit(var) : ~mkLit(var) );
        }
        S.addClause_(lits);
    }
    assert(nvar == S.nVars());
        
    printf("|  Number of variables:  %12d                  |\n", S.nVars());
    printf("|  Number of clauses:    %12d                  |\n", S.nClauses());
        
    if (!S.simplify()) {
        printf("UNSAT\n");
        printf("Solved by unit propagation\n");
        printStats(S);
        printf("\n");
        printf("UNSATISFIABLE\n");
        exit(20);
    }
        
    vec<Lit> dummy;
    lbool ret = S.solveLimited(dummy);
    printStats(S);
    printf("\n");
    printf(ret == l_True ? "SATISFIABLE\n" : ret == l_False ? "UNSATISFIABLE\n" : "INDETERMINATE\n");

    gettimeofday(&tend, NULL);
    int elapsedms = (tend.tv_sec - tstart.tv_sec) * 1000 +
                    tend.tv_usec / 1000 - tstart.tv_usec / 1000;
    printf("time elapsed = %d.%03u s\n", elapsedms/1000, elapsedms%1000);

    return 0;
}

EOF

